#include <algorithm>
#include <cstdio>
#include <list>
#include <typeinfo>
#include <vector>

#include "parser.h"

static const unsigned int MAX_UNDERLINE = 3;

#define DEBUG 1

#define NUM_COLORS 8
char colors[NUM_COLORS] = { 'r', 'g', 'b', 'c', 'm', 'y', 'k', 'w'};

token* ROOT;

void print_result(token& root);
void print_result2(token& root);
void print_tree(token& root, int depth);

void pass_one(token& root);
void remove_dup_trees(token& root);

void merge_colors(token& root);
void merge_colors2(token& root);

void merge_sizes(token& root);
void merge_sizes2(token& root);

void print_result(token& root) {
    token tok;

    for (token_iterator it = root->cbegin(); it != root->cend(); ++it) {
        tok = (*it);

        if (tok->is_tag) {
            cout << '<' << tok->str << '>';
            print_result(tok);
            cout << "</" << tok->str << '>';
        }
        else {
            cout << tok->str;

            if (tok->pl_tok) {
                token pl_token(tok->pl_tok);
                print_result2(pl_token);
            }
        }
    }
}

void print_result2(token& tok) {
    if (tok->is_tag) {
        cout << '<' << tok->str << '>';
        print_result(tok);
        cout << "</" << tok->str << '>';
    }
    else {
        cout << tok->str;
    }
}

void print_context(token_list& tokens) {
    token tok;

    for (token_iterator it = tokens.begin(); it != tokens.end(); ++it) {
        tok = (*it);

        cerr << tok->str << ' ';
        tok->print_context();
        cerr << endl;
    }
}


void expand(token& root, token_list& tokens) {
    token tok;
    token new_tok;
    token prev_tok;

    root->children.clear();

    token_iterator prev_it;

    for (token_iterator it = tokens.begin(); it != tokens.end(); ++it) {
        tok      = (*it);
        prev_it  = it;

        if (it != tokens.begin()) {
            prev_tok = (*--prev_it);

            if (prev_tok->str.size() > 0 &&
                prev_tok->str[prev_tok->str.size() - 1] == ' ' &&
                tok->str.size() > 0 && tok->str[0] == ' ' &&
                prev_tok->context.color == tok->context.color &&
                prev_tok->context.size  == tok->context.size) {

                prev_tok->str.erase(prev_tok->str.size() - 1);
            }
        }

        prev_tok = root;
        for (uint i = 0; i < tok->cons.size(); ++i) {
            new_tok = token(new Token(tok->cons[i], true));
            new_tok->parent = prev_tok();

            prev_tok->children.push_back(new_tok);
            prev_tok = new_tok;
        }
        tok->parent = prev_tok();
        prev_tok->children.push_back(tok);
    }
}

void set_attributes(token& root, Token** prev_sib, int depth) {
    token tok;

    for (token_iterator it = root->cbegin(); it != root->cend(); ++it) {
        tok = (*it);

        tok->depth = depth;

        if (tok->is_tag) {
            set_attributes(tok, prev_sib, depth+1);
        }
        else {
            if (*prev_sib != NULL) {
                tok->prev_sibling = (*prev_sib);
                tok->prev_sibling->next_sibling = tok();
            }
            (*prev_sib) = tok();
        }
    }
}

void pass_one_helper(token& tok, Token* sibling) {
    Token* a;
    Token* b;

    string_list& cons = sibling->cons;
    for (uint i = 0; i < cons.size(); ++i) {
/*         cout << "cons = " << cons[i] << endl; */
        if (tok->contains_context(cons[i])) {
            for (b = sibling->parent; b->str != cons[i];  b = b->parent);
            if (b->depth != 1) {
                for (a = b; a->depth != 1; a = a->parent) {
/*                     cout << "sa = " << a->str << endl;         */
/*                     cout << "sp = " << a->parent->str << endl; */
                    swap(a->str, a->parent->str);
                }
            }

            for (b = tok->parent; b->str != cons[i];  b = b->parent);
            if (b->depth != 1) {
                for (a = b; a->depth != 1; a = a->parent) {
/*                     cout << "a = " << a->str << endl;         */
/*                     cout << "p = " << a->parent->str << endl; */
                    swap(a->str, a->parent->str);
                }

/*                 cout << endl; */
            }
        }
    }
}

// puts neighbor trees in the same depth order
void pass_one(token& root) {
    token tok;

    for (token_iterator it = root->cbegin(); it != root->cend(); ++it) {
        tok = (*it);

        if (tok->is_tag) {
            pass_one(tok);
        }
        else {
/*             cout << " tok = " << tok->str << endl; */
            if (tok->prev_sibling && tok->prev_sibling->depth <= tok->depth) {
                pass_one_helper(tok, tok->prev_sibling);
            }
            else if (tok->next_sibling && tok->next_sibling->depth < tok->depth) {
                pass_one_helper(tok, tok->next_sibling);
            }
        }
    }
}

// removes duplicate neighbor trees
void remove_dup_trees(token& root) {
    token tok;
    token next_tok;

    token_iterator it = root->cbegin();
    token_iterator this_it;

    root->children.push_back(token(new Token("")));

    while (it != root->cend()) {
        tok      = (*it);
        next_tok = (*++it);

        if (tok->is_tag) {
            if (tok->str == next_tok->str) {
                tok->children.insert(tok->cend(),
                                     next_tok->cbegin(),
                                     next_tok->cend());

                this_it = it++;
                root->children.erase(this_it);
            }
            else {
                ++it;
            }

            remove_dup_trees(tok);
        }
    }
    root->children.pop_back();
}

void merge_colors(token& root) {
    merge_colors2(root);
    for (token_iterator it = root->cbegin(); it != root->cend(); ++it) {
        merge_colors(*it);
    }
}

void merge_colors2(token& root) {
    token tok;
    token color_tok;

    for (int i = 0; i < NUM_COLORS; ++i) {
        bool on = false;
        token_iterator this_it;
        token_iterator it = root->cbegin();

        while (it != root->cend()) {
            tok = (*it);
            this_it = it++;

            if (tok->getColor() < 'a' || !tok->is_tag) {
                on = false;
            }
            else if (tok->getColor() == colors[i]) {
                if (on) {
                    color_tok->children.insert(color_tok->cend(),
                                               tok->cbegin(), tok->cend());
                    tok->parent->children.erase(this_it);
                }
                else {
                    on = true;
                    color_tok = tok;
                }
            }
            else if (on) {
                color_tok->children.push_back(tok);
                tok->parent->children.erase(this_it);
            }
        }
    }
}

void merge_sizes(token& root) {
    merge_sizes2(root);
    for (token_iterator it = root->cbegin(); it != root->cend(); ++it) {
        merge_sizes(*it);
    }
}

void merge_sizes2(token& root) {
    token tok;
    token size_tok;

    for (int i = 0; i < 10; ++i) {
        bool on = false;
        token_iterator this_it;
        token_iterator it = root->cbegin();

        while (it != root->cend()) {
            tok = (*it);
            this_it = it++;

            if (tok->getSize() < '0' || tok->getSize() > '9' || !tok->is_tag) {
                on = false;
            }
            else if (tok->getColor() == i + '0') {
                if (on) {
                    size_tok->children.insert(size_tok->cend(),
                                              tok->cbegin(), tok->cend());
                    tok->parent->children.erase(this_it);
                }
                else {
                    on = true;
                    size_tok = tok;
                }
            }
            else if (on) {
                size_tok->children.push_back(tok);
                tok->parent->children.erase(this_it);
            }
        }
    }
}

void fix_pl(token& root, token_list& children, Token* prev_tok, token& pl_tok,
            Token* next_tok) {

    int i = prev_tok->cons.size() - 1;
    int j = next_tok->cons.size() - 1;

    for ( ; i >= 0 && j >= 0; --i, --j) {
        if (prev_tok->cons[i] != next_tok->cons[j]) {
            break;
        }
    }

    // prev_tok and next_tok have all cons in common
    if (i == -1 && j == -1) {
/*         Token* p = next_tok->parent;                                            */
/*         for ( ; p->depth != 1 && !p->isColor() && !p->isSize(); p = p->parent); */
/*                                                                                 */
/*         cout << "prev_tok = " << prev_tok->str << endl; */
/*         cout << "pl_tok   = " << pl_tok->str   << endl; */
/*         cout << "next_tok = " << next_tok->str << endl; */
/*         cout << "parent = " << p->parent->str << endl;  */
/*         token_iterator it = p->parent->children.begin(); */
/*         for ( ; it != p->parent->children.end(); ++it) { */
/*             if ((*it)() == p) {                          */
/*                 p->parent->children.erase(it);           */
/*                 break;                                   */
/*             }                                            */
/*         }                                                */
/*                                                          */
/*         prev_tok->parent->children.push_back(pl_tok);    */
/*         prev_tok->parent->children.push_back(next_tok);  */
        prev_tok->pl_tok = pl_tok();
    }
    else if (i >= 0) {
        prev_tok->pl_tok = pl_tok();
    }
    else if (j >= 0) {
        prev_tok->pl_tok = pl_tok();
    }
}

void optimize_pl(token& root) {
    Token* tok = root->children.front()();
    Token* prev_tok = NULL;
    Token* next_tok = NULL;

    token_iterator it;
    token_list*     children;

    token pl_token;
    token tmp_token;

    // find first leaf
    for (; tok->is_tag; tok = tok->children.front()());

    while (tok != NULL) {
        next_tok = tok->next_sibling;

        if (prev_tok != NULL && tok->next_sibling != NULL &&
            tok->context.pl_able() &&
            (!prev_tok->isColor() && !prev_tok->isSize()) &&
            (!next_tok->isColor() && !next_tok->isSize())) {

            if (prev_tok->good_for_pl(next_tok)) {
                children = &(tok->parent->children);

                for (it = children->begin(); it != children->end(); ++it) {
                    if ((*it)() == tok) {
                        pl_token = token(new Token("PL", true));

                        pl_token->children.push_back(tok);
                        prev_tok->next_sibling = next_tok;
                        next_tok->prev_sibling = prev_tok;
                        children->erase(it);

                        fix_pl(root, *children, prev_tok, pl_token, next_tok);
                        break;
                    }
                }
            }
        }

        prev_tok = tok;
        tok      = tok->next_sibling;
    }
}

void tabs(int depth){
    for (int i = 0; i < depth*4; ++i) {
        cout << ' ';
    }
}

void print_tree(token& root, int depth){
    token tok;

    for (token_iterator it = root->cbegin(); it != root->cend(); ++it) {
        tok = (*it);

        tabs(depth);
        if (tok->is_tag) {
            cout << '<' << tok->str << '>';
            cout << endl;
            print_tree(tok, depth+1);
            tabs(depth);
            cout << "</" << tok->str << '>' << endl;
        }
        else {
            if (tok->str != "\n") {
                cout << tok->str << ' ';
            }
            tok->print_context();
/*             if (tok->prev_sibling) {                       */
/*                 cout << " --> " << tok->prev_sibling->str; */
/*             }                                              */
            cout << endl;
        }
    }
}

int main(int argc, char** argv){
    FILE *in = stdin;

    token root(new Token("root", false));
    token_list tokens;

    ROOT = &root;

    parse(in, tokens);

    if (tokens.size() > 0) {
        expand(root, tokens);

        Token* sib = NULL;
        set_attributes(root, &sib, 1);
        optimize_pl(root);
        merge_colors(root);
        merge_sizes(root);
        pass_one(root);
        remove_dup_trees(root);
    }
    print_result(root);

    cout.flush();
}
